import logging
from typing import AsyncGenerator, List, Optional

from langchain_core.messages.chat import ChatMessage

from khoj.processor.conversation.anthropic.utils import (
    anthropic_chat_completion_with_backoff,
    anthropic_completion_with_backoff,
)
from khoj.processor.conversation.utils import (
    ResponseWithThought,
    messages_to_print,
)

logger = logging.getLogger(__name__)


def anthropic_send_message_to_model(
    messages,
    api_key,
    api_base_url,
    model,
    response_type="text",
    response_schema=None,
    tools=None,
    deepthought=False,
    tracer={},
):
    """
    Send message to model
    """
    # Get response from model. Don't use response_type because Anthropic doesn't support it.
    return anthropic_completion_with_backoff(
        messages=messages,
        system_prompt="",
        model_name=model,
        api_key=api_key,
        api_base_url=api_base_url,
        response_type=response_type,
        response_schema=response_schema,
        tools=tools,
        deepthought=deepthought,
        tracer=tracer,
    )


async def converse_anthropic(
    # Query
    messages: List[ChatMessage],
    # Model
    model: Optional[str] = "claude-3-7-sonnet-latest",
    api_key: Optional[str] = None,
    api_base_url: Optional[str] = None,
    deepthought: Optional[bool] = False,
    tracer: dict = {},
) -> AsyncGenerator[ResponseWithThought, None]:
    """
    Converse with user using Anthropic's Claude
    """
    logger.debug(f"Conversation Context for Claude: {messages_to_print(messages)}")

    # Get Response from Claude
    async for chunk in anthropic_chat_completion_with_backoff(
        messages=messages,
        model_name=model,
        temperature=0.2,
        api_key=api_key,
        api_base_url=api_base_url,
        deepthought=deepthought,
        tracer=tracer,
    ):
        yield chunk



import base64
import io
import logging
import os
import time
from typing import Any, Callable, Dict, List, Optional

import openai
import requests
from google import genai
from google.genai import types as gtypes
from tenacity import (
    retry,
    retry_if_exception,
    retry_if_exception_type,
    stop_after_attempt,
    wait_random_exponential,
)
from tenacity.before_sleep import before_sleep_log

from khoj.database.adapters import ConversationAdapters
from khoj.database.models import (
    Agent,
    ChatMessageModel,
    Intent,
    KhojUser,
    TextToImageModelConfig,
    UserMemory,
)
from khoj.processor.conversation.google.utils import _is_retryable_error
from khoj.processor.conversation.utils import get_image_from_base64, get_image_from_url
from khoj.routers.helpers import ChatEvent, ImageShape, generate_better_image_prompt
from khoj.routers.storage import upload_generated_image_to_bucket
from khoj.utils import state
from khoj.utils.helpers import convert_image_to_webp, is_none_or_empty, timer
from khoj.utils.rawconfig import LocationData

logger = logging.getLogger(__name__)


async def text_to_image(
    message: str,
    user: KhojUser,
    chat_history: List[ChatMessageModel],
    location_data: LocationData,
    references: List[Dict[str, Any]],
    online_results: Dict[str, Any],
    send_status_func: Optional[Callable] = None,
    query_images: Optional[List[str]] = None,
    query_files: str = None,
    relevant_memories: List[UserMemory] = None,
    agent: Agent = None,
    tracer: dict = {},
):
    status_code = 200
    image = None
    image_url = None

    text_to_image_config = await ConversationAdapters.aget_user_text_to_image_model(user)
    if not text_to_image_config:
        # If the user has not configured a text to image model, return an unsupported on server error
        status_code = 501
        message = "Failed to generate image. Setup image generation on the server."
        yield image_url or image, status_code, message
        return

    text2image_model = text_to_image_config.model_name
    image_chat_history: List[ChatMessageModel] = []
    default_intent = Intent(type="remember")
    for chat in chat_history[-4:]:
        if chat.by == "you":
            image_chat_history += [ChatMessageModel(by=chat.by, message=chat.message, intent=default_intent)]
        elif chat.by == "khoj" and chat.images and chat.intent and chat.intent.inferred_queries:
            image_chat_history += [
                ChatMessageModel(by=chat.by, message=chat.intent.inferred_queries[0], intent=default_intent)
            ]
        elif chat.by == "khoj" and chat.intent and chat.intent.type in ["remember", "reminder"]:
            image_chat_history += [ChatMessageModel(by=chat.by, message=chat.message, intent=default_intent)]

    # Generate a better image prompt
    # Use the user's message, chat history, and other context
    if not is_multimodal_model(text2image_model):
        if send_status_func:
            async for event in send_status_func("**Enhancing the Painting Prompt**"):
                yield {ChatEvent.STATUS: event}

        image_prompt_response = await generate_better_image_prompt(
            message,
            image_chat_history,
            location_data=location_data,
            note_references=references,
            online_results=online_results,
            model_type=text_to_image_config.model_type,
            query_images=query_images,
            query_files=query_files,
            relevant_memories=relevant_memories,
            user=user,
            agent=agent,
            tracer=tracer,
        )
        image_prompt = image_prompt_response["description"]
        image_shape = image_prompt_response["shape"]
    else:
        image_prompt = message
        image_shape = None

    if send_status_func:
        async for event in send_status_func(f"**Painting to Imagine**:\n{image_prompt}"):
            yield {ChatEvent.STATUS: event}

    # Generate image using the configured model and API
    with timer(f"Generate image with {text_to_image_config.model_type}", logger):
        try:
            if text_to_image_config.model_type == TextToImageModelConfig.ModelType.OPENAI:
                webp_image_bytes = generate_image_with_openai(
                    image_prompt, text_to_image_config, text2image_model, image_shape
                )
            elif text_to_image_config.model_type == TextToImageModelConfig.ModelType.REPLICATE:
                webp_image_bytes = generate_image_with_replicate(
                    image_prompt, text_to_image_config, text2image_model, image_shape
                )
            elif text_to_image_config.model_type == TextToImageModelConfig.ModelType.GOOGLE:
                webp_image_bytes = generate_image_with_google(
                    image_prompt,
                    text_to_image_config,
                    text2image_model,
                    image_shape,
                    chat_history=chat_history,
                    query_images=query_images,
                )
        except openai.OpenAIError or openai.BadRequestError or openai.APIConnectionError as e:
            if "content_policy_violation" in e.message:
                logger.error(f"Image Generation blocked by OpenAI: {e}")
                status_code = e.status_code  # type: ignore
                message = "Image generation blocked by OpenAI due to policy violation"  # type: ignore
                yield image_url or image, status_code, message
                return
            else:
                logger.error(f"Image Generation failed with {e}", exc_info=True)
                message = "Image generation failed using OpenAI"  # type: ignore
                status_code = e.status_code  # type: ignore
                yield image_url or image, status_code, message
                return
        except ValueError as e:
            logger.error(f"Image Generation failed with {e}", exc_info=True)
            message = f"Image generation using {text2image_model} via {text_to_image_config.model_type} failed due to an unknown error"
            status_code = 500
            yield image_url or image, status_code, message
            return
        except requests.RequestException as e:
            logger.error(f"Image Generation failed with {e}", exc_info=True)
            message = f"Image generation using {text2image_model} via {text_to_image_config.model_type} failed due to a network error."
            status_code = 502
            yield image_url or image, status_code, message
            return

    # Decide how to store the generated image
    with timer("Upload image to S3", logger):
        image_url = upload_generated_image_to_bucket(webp_image_bytes, user.uuid)

    if not image_url:
        image = f"data:image/webp;base64,{base64.b64encode(webp_image_bytes).decode('utf-8')}"

    yield image_url or image, status_code, image_prompt


@retry(
    retry=(
        retry_if_exception_type(openai.APITimeoutError)
        | retry_if_exception_type(openai.APIError)
        | retry_if_exception_type(openai.APIConnectionError)
        | retry_if_exception_type(openai.RateLimitError)
        | retry_if_exception_type(openai.APIStatusError)
    ),
    wait=wait_random_exponential(min=1, max=10),
    stop=stop_after_attempt(3),
    before_sleep=before_sleep_log(logger, logging.DEBUG),
    reraise=True,
)
def generate_image_with_openai(
    improved_image_prompt: str,
    text_to_image_config: TextToImageModelConfig,
    text2image_model: str,
    shape: ImageShape = ImageShape.SQUARE,
):
    "Generate image using OpenAI (compatible) API"

    # Get the API config from the user's configuration
    api_key = None
    if text_to_image_config.api_key:
        api_key = text_to_image_config.api_key
        openai_client = openai.OpenAI(api_key=api_key)
    elif text_to_image_config.ai_model_api:
        api_key = text_to_image_config.ai_model_api.api_key
        api_base_url = text_to_image_config.ai_model_api.api_base_url
        openai_client = openai.OpenAI(api_key=api_key, base_url=api_base_url)
    elif state.openai_client:
        openai_client = state.openai_client

    # Convert shape to size for OpenAI
    if shape == ImageShape.PORTRAIT:
        size = "1024x1536"
    elif shape == ImageShape.LANDSCAPE:
        size = "1536x1024"
    else:  # Square
        size = "1024x1024"

    # Generate image using OpenAI API
    OPENAI_IMAGE_GEN_STYLE = "vivid"
    response = openai_client.images.generate(
        prompt=improved_image_prompt,
        model=text2image_model,
        style=OPENAI_IMAGE_GEN_STYLE,
        size=size,
        response_format="b64_json",
    )

    # Extract the base64 image from the response
    image = response.data[0].b64_json
    # Decode base64 png and convert it to webp for faster loading
    return convert_image_to_webp(base64.b64decode(image))


@retry(
    retry=retry_if_exception_type(requests.RequestException),
    wait=wait_random_exponential(min=1, max=10),
    stop=stop_after_attempt(3),
    before_sleep=before_sleep_log(logger, logging.DEBUG),
    reraise=True,
)
def generate_image_with_replicate(
    improved_image_prompt: str,
    text_to_image_config: TextToImageModelConfig,
    text2image_model: str,
    shape: ImageShape = ImageShape.SQUARE,
):
    "Generate image using Replicate API"

    # Convert shape to aspect ratio for Replicate
    # Replicate supports only 1:1, 3:4, and 4:3 aspect ratios
    if shape == ImageShape.PORTRAIT:
        aspect_ratio = "3:4"
    elif shape == ImageShape.LANDSCAPE:
        aspect_ratio = "4:3"
    else:  # Square
        aspect_ratio = "1:1"

    # Create image generation task on Replicate
    replicate_create_prediction_url = f"https://api.replicate.com/v1/models/{text2image_model}/predictions"
    headers = {
        "Authorization": f"Bearer {text_to_image_config.api_key}",
        "Content-Type": "application/json",
    }
    json = {
        "input": {
            "prompt": improved_image_prompt,
            "num_outputs": 1,
            "aspect_ratio": aspect_ratio,
            "output_format": "webp",
            "output_quality": 100,
        }
    }

    seed = int(os.getenv("KHOJ_LLM_SEED")) if os.getenv("KHOJ_LLM_SEED") else None
    if seed:
        json["input"]["seed"] = seed

    create_prediction = requests.post(replicate_create_prediction_url, headers=headers, json=json).json()

    # Get status of image generation task
    get_prediction_url = create_prediction["urls"]["get"]
    get_prediction = requests.get(get_prediction_url, headers=headers).json()
    status = get_prediction["status"]
    retry_count = 1

    # Poll the image generation task for completion status
    while status not in ["succeeded", "failed", "canceled"] and retry_count < 20:
        time.sleep(2)
        get_prediction = requests.get(get_prediction_url, headers=headers).json()
        status = get_prediction["status"]
        retry_count += 1

    # Raise exception if the image generation task fails
    if status != "succeeded":
        error = get_prediction.get("error")
        if retry_count >= 10:
            raise requests.RequestException("Image generation timed out")
        raise requests.RequestException(f"Image generation failed with status: {status}, message: {error}")

    # Get the generated image
    image_url = get_prediction["output"][0] if isinstance(get_prediction["output"], list) else get_prediction["output"]
    return io.BytesIO(requests.get(image_url).content).getvalue()


@retry(
    retry=retry_if_exception(_is_retryable_error),
    wait=wait_random_exponential(min=1, max=10),
    stop=stop_after_attempt(3),
    before_sleep=before_sleep_log(logger, logging.DEBUG),
    reraise=True,
)
def generate_image_with_google(
    improved_image_prompt: str,
    text_to_image_config: TextToImageModelConfig,
    text2image_model: str,
    shape: ImageShape = ImageShape.SQUARE,
    chat_history: List[ChatMessageModel] = [],
    query_images: List[str] = [],
):
    """Generate image using Google's AI over API"""

    # Initialize the Google AI client
    api_key = text_to_image_config.api_key or text_to_image_config.ai_model_api.api_key
    client = genai.Client(api_key=api_key)

    # Convert shape to aspect ratio for Google
    if shape == ImageShape.PORTRAIT:
        aspect_ratio = "3:4"
    elif shape == ImageShape.LANDSCAPE:
        aspect_ratio = "4:3"
    else:  # Square
        aspect_ratio = "1:1"

    image_bytes = None
    if is_multimodal_model(text2image_model):
        # Format chat history for Gemini
        contents = format_messages_for_gemini(improved_image_prompt, text2image_model, chat_history, query_images)

        # Configure image generation settings
        image_size = "2K" if text2image_model.startswith("gemini-3") else None
        config = gtypes.GenerateContentConfig(
            response_modalities=["IMAGE"], image_config=gtypes.ImageConfig(aspect_ratio=None, image_size=image_size)
        )

        # Call the Gemini API to generate the image
        response = client.models.generate_content(
            contents=contents,
            model=text2image_model,
            config=config,
        )

        # Extract the image bytes from the first generated image
        for part in response.parts or []:
            if part.inline_data is not None:
                image = part.as_image()
                image_bytes = image.image_bytes
                break
        if not image_bytes:
            raise ValueError("Failed to generate image using Google AI")
    else:
        # Configure image generation settings
        config = gtypes.GenerateImagesConfig(
            number_of_images=1,
            safety_filter_level=gtypes.SafetyFilterLevel.BLOCK_LOW_AND_ABOVE,
            person_generation=gtypes.PersonGeneration.ALLOW_ADULT,
            include_rai_reason=True,
            output_mime_type="image/png",
            aspect_ratio=aspect_ratio,
        )

        # Call the Gemini API to generate the image
        response = client.models.generate_images(model=text2image_model, prompt=improved_image_prompt, config=config)

        if not response.generated_images:
            raise ValueError("Failed to generate image using Google AI")

        # Extract the image bytes from the first generated image
        image_bytes = response.generated_images[0].image.image_bytes

    # Convert to webp for faster loading
    return convert_image_to_webp(image_bytes)


def format_messages_for_gemini(
    improved_image_prompt: str,
    text2image_model: str,
    chat_history: List[ChatMessageModel] = [],
    query_images: List[str] = [],
) -> List[gtypes.Content]:
    """Format chat messages for Gemini multimodal models.

    Reframes assistant messages with generated images as user messages to enable
    multi-turn image editing with gemini 3 models.
    """
    contents = []
    for chat in chat_history:
        role = "model" if chat.by == "khoj" else "user"
        parts = []

        # Reframe assistant messages to gemini 3 as user messages
        # This enables multi-turn image edits without storing, passing thought_signature required by gemini 3 models
        if role == "model" and text2image_model.startswith("gemini-3"):
            if chat.images:
                parts.append(gtypes.Part.from_text(text="This is the image you previously generated:"))
                for image_data in chat.images:
                    if image_data.startswith("http"):
                        image = get_image_from_url(image_data, type="bytes")
                    else:
                        image = get_image_from_base64(image_data, type="bytes")
                parts.append(gtypes.Part.from_bytes(data=image.content, mime_type=image.type))
            else:
                parts.append(gtypes.Part.from_text(text="This is the message you previously sent:"))
                messages = chat.message if isinstance(chat.message, list) else [chat.message]  # type: ignore[list-item]
                for text in messages:
                    if isinstance(text, dict) and not is_none_or_empty(text.get("text")):
                        parts.append(gtypes.Part.from_text(text=text.get("text")))
                    elif isinstance(text, str):
                        parts.append(gtypes.Part.from_text(text=text))
            contents.append(gtypes.Content(role="user", parts=parts))
            continue

        # Handle regular messages
        for image_data in chat.images or []:
            if image_data.startswith("http"):
                image = get_image_from_url(image_data, type="bytes")
            else:
                image = get_image_from_base64(image_data, type="bytes")
            parts.append(gtypes.Part.from_bytes(data=image.content, mime_type=image.type))
        messages = chat.message if isinstance(chat.message, list) else [chat.message]  # type: ignore[list-item]
        for text in messages:
            if isinstance(text, dict) and not is_none_or_empty(text.get("text")):
                parts.append(gtypes.Part.from_text(text=text.get("text")))
            elif isinstance(text, str):
                parts.append(gtypes.Part.from_text(text=text))
        contents.append(gtypes.Content(role=role, parts=parts))

    query_parts = []
    for img in query_images or []:
        if img.startswith("http"):
            image = get_image_from_url(img, type="bytes")
        else:
            image = get_image_from_base64(img, type="bytes")
        query_parts.append(gtypes.Part.from_bytes(data=image.content, mime_type=image.type))
    query_parts.append(gtypes.Part.from_text(text=improved_image_prompt))

    contents += [gtypes.Content(role="user", parts=query_parts)]
    return contents


def is_multimodal_model(model_name: str) -> bool:
    """Check if the model can see and generate images"""
    multimodal_models = ["gemini-2.5-flash-image", "gemini-3-pro-image-preview"]
    return model_name.lower() in multimodal_models



# Generated by Django 5.1.14 on 2025-11-30 00:47

from django.db import migrations, models


def delete_stability_configs(apps, schema_editor):
    TextToImageModelConfig = apps.get_model("database", "TextToImageModelConfig")
    TextToImageModelConfig.objects.filter(model_type="stability-ai").delete()


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0097_serverchatsettings_priority"),
    ]

    operations = [
        migrations.RunPython(delete_stability_configs, migrations.RunPython.noop),
        migrations.AlterField(
            model_name="texttoimagemodelconfig",
            name="model_type",
            field=models.CharField(
                choices=[
                    ("openai", "Openai"),
                    ("replicate", "Replicate"),
                    ("google", "Google"),
                ],
                default="openai",
                max_length=200,
            ),
        ),
    ]



import logging
from typing import Any, AsyncGenerator, Dict, List, Optional

from langchain_core.messages.chat import ChatMessage

from khoj.processor.conversation.openai.utils import (
    chat_completion_with_backoff,
    clean_response_schema,
    completion_with_backoff,
    get_structured_output_support,
    is_cerebras_api,
    responses_chat_completion_with_backoff,
    responses_completion_with_backoff,
    supports_responses_api,
    to_openai_tools,
)
from khoj.processor.conversation.utils import (
    ResponseWithThought,
    StructuredOutputSupport,
    messages_to_print,
)
from khoj.utils.helpers import ToolDefinition

logger = logging.getLogger(__name__)


def openai_send_message_to_model(
    messages,
    api_key,
    model: str,
    response_type="text",
    response_schema=None,
    tools: list[ToolDefinition] = None,
    deepthought=False,
    api_base_url: str | None = None,
    tracer: dict = {},
):
    """
    Send message to model
    """

    model_kwargs: Dict[str, Any] = {}
    json_support = get_structured_output_support(model, api_base_url)
    strict = not is_cerebras_api(api_base_url)
    if tools and json_support == StructuredOutputSupport.TOOL:
        model_kwargs["tools"] = to_openai_tools(tools, model=model, api_base_url=api_base_url)
    elif response_schema and json_support >= StructuredOutputSupport.SCHEMA:
        # Drop unsupported fields from schema passed to OpenAI APi
        cleaned_response_schema = clean_response_schema(response_schema)
        if supports_responses_api(model, api_base_url):
            model_kwargs["text"] = {
                "format": {
                    "type": "json_schema",
                    "strict": strict,
                    "name": response_schema.__name__,
                    "schema": cleaned_response_schema,
                }
            }
        else:
            model_kwargs["response_format"] = {
                "type": "json_schema",
                "json_schema": {
                    "schema": cleaned_response_schema,
                    "name": response_schema.__name__,
                    "strict": strict,
                },
            }
    elif response_type == "json_object" and json_support == StructuredOutputSupport.OBJECT:
        model_kwargs["response_format"] = {"type": response_type}

    # Get Response from GPT
    if supports_responses_api(model, api_base_url):
        return responses_completion_with_backoff(
            messages=messages,
            model_name=model,
            openai_api_key=api_key,
            api_base_url=api_base_url,
            deepthought=deepthought,
            model_kwargs=model_kwargs,
            tracer=tracer,
        )
    else:
        return completion_with_backoff(
            messages=messages,
            model_name=model,
            openai_api_key=api_key,
            api_base_url=api_base_url,
            deepthought=deepthought,
            model_kwargs=model_kwargs,
            tracer=tracer,
        )


async def converse_openai(
    # Query
    messages: List[ChatMessage],
    # Model
    model: str = "gpt-4.1-mini",
    api_key: Optional[str] = None,
    api_base_url: Optional[str] = None,
    temperature: float = 0.6,
    deepthought: Optional[bool] = False,
    tracer: dict = {},
) -> AsyncGenerator[ResponseWithThought, None]:
    """
    Converse with user using OpenAI's ChatGPT
    """
    logger.debug(f"Conversation Context for GPT: {messages_to_print(messages)}")

    # Get Response from GPT
    if supports_responses_api(model, api_base_url):
        async for chunk in responses_chat_completion_with_backoff(
            messages=messages,
            model_name=model,
            temperature=temperature,
            openai_api_key=api_key,
            api_base_url=api_base_url,
            deepthought=deepthought,
            tracer=tracer,
        ):
            yield chunk
    else:
        # For non-OpenAI APIs, use the chat completion method
        async for chunk in chat_completion_with_backoff(
            messages=messages,
            model_name=model,
            temperature=temperature,
            openai_api_key=api_key,
            api_base_url=api_base_url,
            deepthought=deepthought,
            tracer=tracer,
        ):
            yield chunk



import json
import logging
import math
import os
import uuid
from typing import List, Optional, Union

import openai
from fastapi import APIRouter, Depends, File, HTTPException, Request, UploadFile
from fastapi.responses import Response
from starlette.authentication import has_required_scope, requires

from khoj.configure import initialize_content
from khoj.database import adapters
from khoj.database.adapters import ConversationAdapters, EntryAdapters, get_user_photo
from khoj.database.models import KhojUser, SpeechToTextModelOptions, UserConversationConfig
from khoj.processor.conversation.openai.whisper import transcribe_audio
from khoj.routers.helpers import (
    ApiUserRateLimiter,
    CommonQueryParams,
    ConversationCommandRateLimiter,
    execute_search,
    get_user_config,
    update_telemetry_state,
)
from khoj.utils import state
from khoj.utils.rawconfig import SearchResponse
from khoj.utils.state import SearchType

# Initialize Router
api = APIRouter()
logger = logging.getLogger(__name__)
conversation_command_rate_limiter = ConversationCommandRateLimiter(
    trial_rate_limit=2, subscribed_rate_limit=100, slug="command"
)


@api.delete("/self")
@requires(["authenticated"])
def delete_self(request: Request):
    user = request.user.object
    user.delete()
    return {"status": "ok"}


@api.get("/search", response_model=List[SearchResponse])
@requires(["authenticated"])
async def search(
    q: str,
    request: Request,
    common: CommonQueryParams,
    n: Optional[int] = 5,
    t: Optional[SearchType] = SearchType.All,
    r: Optional[bool] = False,
    max_distance: Optional[Union[float, None]] = None,
    dedupe: Optional[bool] = True,
):
    user = request.user.object

    results = await execute_search(
        user=user,
        q=q,
        n=n,
        t=t,
        r=r,
        max_distance=max_distance or math.inf,
        dedupe=dedupe,
    )

    update_telemetry_state(
        request=request,
        telemetry_type="api",
        api="search",
        **common.__dict__,
    )

    return results


@api.get("/update")
@requires(["authenticated"])
def update(
    request: Request,
    common: CommonQueryParams,
    t: Optional[SearchType] = None,
    force: Optional[bool] = False,
):
    user = request.user.object
    try:
        initialize_content(user=user, regenerate=force, search_type=t)
    except Exception as e:
        error_msg = f"🚨 Failed to update server indexed content via API: {e}"
        logger.error(error_msg, exc_info=True)
        raise HTTPException(status_code=500, detail=error_msg)
    else:
        logger.info("📪 Server indexed content updated via API")

    update_telemetry_state(
        request=request,
        telemetry_type="api",
        api="update",
        **common.__dict__,
    )

    return {"status": "ok", "message": "khoj reloaded"}


@api.post("/transcribe")
@requires(["authenticated"])
async def transcribe(
    request: Request,
    common: CommonQueryParams,
    file: UploadFile = File(...),
    rate_limiter_per_minute=Depends(
        ApiUserRateLimiter(requests=20, subscribed_requests=20, window=60, slug="transcribe_minute")
    ),
    rate_limiter_per_day=Depends(
        ApiUserRateLimiter(requests=60, subscribed_requests=600, window=60 * 60 * 24, slug="transcribe_day")
    ),
):
    user: KhojUser = request.user.object
    audio_filename = f"{user.uuid}-{str(uuid.uuid4())}.webm"
    user_message: str = None

    # If the file is too large, return an unprocessable entity error
    if file.size > 10 * 1024 * 1024:
        logger.warning(f"Audio file too large to transcribe. Audio file size: {file.size}. Exceeds 10Mb limit.")
        return Response(content="Audio size larger than 10Mb limit", status_code=422)

    # Transcribe the audio from the request
    try:
        # Store the audio from the request in a temporary file
        audio_data = await file.read()
        with open(audio_filename, "wb") as audio_file_writer:
            audio_file_writer.write(audio_data)
        audio_file = open(audio_filename, "rb")

        # Send the audio data to the Whisper API
        speech_to_text_config = await ConversationAdapters.get_speech_to_text_config()
        if not speech_to_text_config:
            # If the user has not configured a speech to text model, return an unsupported on server error
            status_code = 501
        elif speech_to_text_config.model_type == SpeechToTextModelOptions.ModelType.OPENAI:
            speech2text_model = speech_to_text_config.model_name
            if speech_to_text_config.ai_model_api:
                api_key = speech_to_text_config.ai_model_api.api_key
                api_base_url = speech_to_text_config.ai_model_api.api_base_url
                openai_client = openai.OpenAI(api_key=api_key, base_url=api_base_url)
            elif state.openai_client:
                openai_client = state.openai_client
            if openai_client:
                user_message = await transcribe_audio(audio_file, speech2text_model, client=openai_client)
            else:
                status_code = 501
    finally:
        # Close and Delete the temporary audio file
        audio_file.close()
        os.remove(audio_filename)

    if user_message is None:
        return Response(status_code=status_code or 500)

    update_telemetry_state(
        request=request,
        telemetry_type="api",
        api="transcribe",
        **common.__dict__,
    )

    # Return the spoken text
    content = json.dumps({"text": user_message})
    return Response(content=content, media_type="application/json", status_code=200)


@api.get("/settings", response_class=Response)
@requires(["authenticated"])
def get_settings(request: Request, detailed: Optional[bool] = False) -> Response:
    user = request.user.object
    user_config = get_user_config(user, request, is_detailed=detailed)
    del user_config["request"]

    # Return config data as a JSON response
    return Response(content=json.dumps(user_config), media_type="application/json", status_code=200)


@api.patch("/user/name", status_code=200)
@requires(["authenticated"])
def set_user_name(
    request: Request,
    name: str,
    client: Optional[str] = None,
):
    user = request.user.object

    split_name = name.split(" ")

    if len(split_name) > 2:
        raise HTTPException(status_code=400, detail="Name must be in the format: Firstname Lastname")

    if len(split_name) == 1:
        first_name = split_name[0]
        last_name = ""
    else:
        first_name, last_name = split_name[0], split_name[-1]

    adapters.set_user_name(user, first_name, last_name)

    update_telemetry_state(
        request=request,
        telemetry_type="api",
        api="set_user_name",
        client=client,
    )

    return {"status": "ok"}


@api.patch("/user/memory", status_code=200)
@requires(["authenticated"])
def set_user_memory_enabled(
    request: Request,
    enable_memory: bool,
    client: Optional[str] = None,
):
    user = request.user.object

    user_config, _ = UserConversationConfig.objects.get_or_create(user=user)
    user_config.enable_memory = enable_memory
    user_config.save()

    update_telemetry_state(
        request=request,
        telemetry_type="api",
        api="set_user_memory_enabled",
        client=client,
    )

    return {"status": "ok", "enable_memory": enable_memory}


@api.get("/health", response_class=Response)
@requires(["authenticated"], status_code=200)
def health_check(request: Request) -> Response:
    response_obj = {"email": request.user.object.email}
    return Response(content=json.dumps(response_obj), media_type="application/json", status_code=200)


@api.get("/v1/user", response_class=Response)
@requires(["authenticated"])
def user_info(request: Request) -> Response:
    # Get user information
    user: KhojUser = request.user.object
    user_picture = get_user_photo(user=user)
    is_active = has_required_scope(request, ["premium"])
    has_documents = EntryAdapters.user_has_entries(user=user)

    # Collect user information in a dictionary
    user_info = {
        "email": user.email,
        "username": user.username,
        "photo": user_picture,
        "is_active": is_active,
        "has_documents": has_documents,
        "khoj_version": state.khoj_version,
    }

    # Return user information as a JSON response
    return Response(content=json.dumps(user_info), media_type="application/json", status_code=200)



import logging
from typing import AsyncGenerator, List, Optional

from langchain_core.messages.chat import ChatMessage

from khoj.processor.conversation.google.utils import (
    gemini_chat_completion_with_backoff,
    gemini_completion_with_backoff,
)
from khoj.processor.conversation.utils import (
    ResponseWithThought,
    messages_to_print,
)

logger = logging.getLogger(__name__)


def gemini_send_message_to_model(
    messages,
    api_key,
    model,
    api_base_url=None,
    response_type="text",
    response_schema=None,
    tools=None,
    model_kwargs=None,
    deepthought=False,
    tracer={},
):
    """
    Send message to model
    """
    model_kwargs = {}

    if tools:
        model_kwargs["tools"] = tools
    # Monitor for flakiness in 1.5+ models. This would cause unwanted behavior and terminate response early in 1.5 models.
    elif response_type == "json_object":
        model_kwargs["response_mime_type"] = "application/json"
        if response_schema:
            model_kwargs["response_schema"] = response_schema

    # Get Response from Gemini
    return gemini_completion_with_backoff(
        messages=messages,
        system_prompt="",
        model_name=model,
        api_key=api_key,
        api_base_url=api_base_url,
        model_kwargs=model_kwargs,
        deepthought=deepthought,
        tracer=tracer,
    )


async def converse_gemini(
    # Query
    messages: List[ChatMessage],
    # Model
    model: Optional[str] = "gemini-2.5-flash",
    api_key: Optional[str] = None,
    api_base_url: Optional[str] = None,
    temperature: float = 1.0,
    deepthought: Optional[bool] = False,
    tracer={},
) -> AsyncGenerator[ResponseWithThought, None]:
    """
    Converse with user using Google's Gemini
    """
    logger.debug(f"Conversation Context for Gemini: {messages_to_print(messages)}")

    # Get Response from Google AI
    async for chunk in gemini_chat_completion_with_backoff(
        messages=messages,
        model_name=model,
        temperature=temperature,
        api_key=api_key,
        api_base_url=api_base_url,
        deepthought=deepthought,
        tracer=tracer,
    ):
        yield chunk



# Generated by Django 5.1.8 on 2025-04-18 15:15

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0088_ratelimitrecord"),
    ]

    operations = [
        migrations.AddField(
            model_name="chatmodel",
            name="price_tier",
            field=models.CharField(choices=[("free", "Free"), ("standard", "Standard")], default="free", max_length=20),
        ),
        migrations.AddField(
            model_name="speechtotextmodeloptions",
            name="price_tier",
            field=models.CharField(choices=[("free", "Free"), ("standard", "Standard")], default="free", max_length=20),
        ),
        migrations.AddField(
            model_name="texttoimagemodelconfig",
            name="price_tier",
            field=models.CharField(choices=[("free", "Free"), ("standard", "Standard")], default="free", max_length=20),
        ),
        migrations.AddField(
            model_name="voicemodeloption",
            name="price_tier",
            field=models.CharField(
                choices=[("free", "Free"), ("standard", "Standard")], default="standard", max_length=20
            ),
        ),
    ]



import json
import logging
from typing import Optional

from asgiref.sync import sync_to_async
from fastapi import APIRouter, Request
from fastapi.responses import Response
from pydantic import BaseModel
from starlette.authentication import requires

from khoj.database.adapters import UserMemoryAdapters
from khoj.database.models import UserMemory

api_memories = APIRouter()
logger = logging.getLogger(__name__)


@api_memories.get("")
@requires(["authenticated"])
async def get_memories(
    request: Request,
    client: Optional[str] = None,
):
    """Get all memories for the authenticated user"""
    user = request.user.object

    memories = UserMemory.objects.filter(user=user)
    all_memories = await sync_to_async(list)(memories)

    # Convert memories to a list of dictionaries
    formatted_memories = [
        {
            "id": memory.id,
            "raw": memory.raw,
            "created_at": memory.created_at.isoformat(),
        }
        for memory in all_memories
    ]

    return Response(content=json.dumps(formatted_memories), media_type="application/json", status_code=200)


@api_memories.delete("/{memory_id}")
@requires(["authenticated"])
async def delete_memory(
    request: Request,
    memory_id: int,
    client: Optional[str] = None,
):
    """Delete a specific memory by ID"""
    user = request.user.object

    # Verify memory belongs to user before deleting
    memory = await UserMemory.objects.filter(id=memory_id, user=user).afirst()
    if not memory:
        return Response(
            content=json.dumps({"error": "Memory not found"}), media_type="application/json", status_code=404
        )

    await memory.adelete()

    return Response(status_code=204)


class UpdateMemoryBody(BaseModel):
    """Request model for updating a memory"""

    raw: str


@api_memories.put("/{memory_id}")
@requires(["authenticated"])
async def update_memory(
    request: Request,
    body: UpdateMemoryBody,
    memory_id: int,
    client: Optional[str] = None,
):
    """Update a specific memory's content"""
    user = request.user.object

    # Get the memory and verify it belongs to the user
    memory = await UserMemory.objects.filter(id=memory_id, user=user).afirst()
    if not memory:
        return Response(
            content=json.dumps({"error": "Memory not found"}), media_type="application/json", status_code=404
        )

    new_content = body.raw
    if not new_content:
        return Response(
            content=json.dumps({"error": "Missing required field 'raw'"}),
            media_type="application/json",
            status_code=400,
        )

    await memory.adelete()

    # Create a new memory with the updated content
    new_memory = await UserMemoryAdapters.save_memory(
        user=user,
        memory=new_content,
    )

    return Response(
        content=json.dumps(
            {
                "id": new_memory.id,
                "raw": new_memory.raw,
            }
        ),
        media_type="application/json",
        status_code=200,
    )
